#pragma once
#include "TypeLists.h"
#include <vector>
#include <memory>

namespace RegExp {

    using namespace std;

    enum ERROR_STATE {
        esNoError            = 0,
        esLimitError         = 1,
        esNoMatch            = 2,
        esOtherError         = 3
    };
    enum CONSTS {
        MAX_STR_SIZE         = 100
    };
    template <class A, class B>
    inline A min(A a, B b) { return (a<(A)b)?a:(A)b; }

    //   
    //////////////////////////////////////////////////////////////////////////////////////////////
    class CResIf
    { public:
        virtual ~CResIf() {}
        virtual void AddNode(CResIf* n) = 0;
        virtual void AddValue(const char* v, unsigned size = 0) = 0;
        virtual unsigned GetType(void) const = 0;
        virtual void SetParent(CResIf* p) {}
    };

    //    
    //////////////////////////////////////////////////////////////////////////////////////////////
    // Match    .   src==NULL,    
    //          
    // Reset   ,     
    class CRegExpIf
    { public:
        CRegExpIf(void): Catches() {}
        virtual bool           Match(CResIf*& r, const char* src = NULL, unsigned size = 0) = 0;
        virtual ERROR_STATE  Replace(char* dst, unsigned size) = 0;
        virtual CRegExpIf*     Clone(void) const = 0;
        virtual const char* Catch(unsigned level, unsigned& size) 
        { if ((level==0)||(level>Catches.size())) return NULL;
          size = Catches[level-1].Size;
          return Catches[level-1].Start;
        }
        virtual void Reset(void)
        { Catches.clear();
        }
        virtual void Save(const char* Start, unsigned Size)
        { Catches.push_back(CCatch(Start,Size));
        }
        virtual void Forget(const char* Start)
        { for (CIter p=Catches.begin();p!=Catches.end();++p)
               if (p->Start>=Start) {
                   Catches.erase(p);
                   break;
               }
        }
	  private:
		struct CCatch {
          CCatch(const char* start, unsigned size): Start(start), Size(size) {}
		  const char*  Start;
		  unsigned     Size;
		};
        vector<CCatch> Catches;
        typedef vector<CCatch>::iterator CIter;
    };

    enum PREDICATE_TYPE {
        ptSpace              = 0,       // \s
        ptBegin              = 1,       // ^
        ptEnd                = 2,       // $
        ptSomeChar           = 3,       // .
        ptDigit              = 4,       // \d
        ptWordChar           = 5,       // \w
        ptWordBorder         = 6        // \b
    };

    enum QUANTIFIER_TYPE {
        qtGreed              = 0,
        qtLazy               = 1,
        qtFast               = 2
    };

    //  
    //////////////////////////////////////////////////////////////////////////////////////////////
    class O 
    { public:
        O(): IsOnce(true) {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { if (!IsOnce) return -1;
          IsOnce = false;          
          return size;
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) { return esOtherError; }
        void Clear(void) { IsOnce = true; }
      private:
        bool IsOnce;
    };

    //  
    //////////////////////////////////////////////////////////////////////////////////////////////
    //      
    //
    // 1.  Head.Match,    -1 (  )
    // 2.  ,  Clear  Head  Tail
    template <class T> class E;
    template <> class E<NullType> 
    { public:
       int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) {return 0;} 
	   ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) {return esNoError;}
       void Clear(void) {}
    };
    template <class T, class U>
    class E< TypeList<T,U> > 
    { public:
        E(): Head(), Tail() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { int Offset = 0;
          for (;;) {
              Offset = Head.Match(m,s,size,res);
              if (Offset<0) {
                  Clear();
                  return Offset;
              }
              int r = Tail.Match(m,s+Offset,size-Offset,res);
              if (r>=0) {
                  Offset += r;
                  break;
              }
          }
          return Offset;
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) 
        { ERROR_STATE r = Head.Replace(m,dst,maxsize);
          if (r!=esNoError) return r;
          return Tail.Replace(m,dst,maxsize);
        }
        void Clear(void) 
        { Head.Clear();
          Tail.Clear();
        }
      private:
        T      Head;
        E< U > Tail;
    };
    template <class T> class E
    { public:
        E(): Elem() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { return Elem.Match(m,s,size,res);
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) 
        { return Elem.Replace(m,dst,maxsize);
        }
        void Clear(void) 
        { Elem.Clear();
        }
      private:
        E< TL_1(T) > Elem;
    };

    //  
    //////////////////////////////////////////////////////////////////////////////////////////////
    //      
    template <class T> class D;
    template <> class D<NullType> 
	{ public:
       int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) {return -1;} 
       void Clear(void) {}
	};
    template <class T, class U>
    class D< TypeList<T,U> > 
	{ public:
	    D(): Head(), Tail() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { int r = Head.Match(m,s,size,res);
		  if (r<0)
			  r = Tail.Match(m,s,size,res);
          return r;
        }
        void Clear(void) 
        { Head.Clear();
          Tail.Clear();
        }
	  private:
        T      Head;
        D< U > Tail;
	};
    template <class T> class D
    { public:
        D(): Elem() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { int r = Elem.Match(m,s,size,res);
		  if (r<0) return 0;
          return r;
        }
        void Clear(void) 
        { Elem.Clear();
        }
      private:
        T Elem;
    };

    // 
    //////////////////////////////////////////////////////////////////////////////////////////////
    //  r==0  Elem.Match       
    //  
    template <class T, class U = NullType, int mn = 0, int mx = 0, QUANTIFIER_TYPE t = qtFast>
    class Q
    { public:
        Q(): Elem(), Stop(), Cnt(-1) {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { return Match(m,s,size,res,Int2Type<t>());
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) 
        { for (int i=0;i<mn;i++) {
               ERROR_STATE r = Elem.Replace(m,dst,maxsize);
               if (r!=esNoError) return r;
          }
          return esNoError;
        }
        void Clear(void) 
        { Elem.Clear();
        }
      private:
        template <class S>
        int StopCond(CRegExpIf* m, const char* s, unsigned size, Type2Type<S>)
        { CResIf* r = NULL;
		  Stop.Clear();
          return Stop.Match(m,s,size,r);
        }
        template <>
        int StopCond(CRegExpIf* m, const char* s, unsigned size, Type2Type<NullType>)
        { return -1;
        }
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res, Int2Type<qtGreed>)
        { unsigned Offset = 0;
          if (Cnt<0) {
              for (Cnt=0;;) {
                   int r = StopCond(m,s+Offset,size-Offset,Type2Type<U>());
                   if (r>=0) break;
                   Elem.Clear();
                   r = Elem.Match(m,s+Offset,size-Offset,res);
                   if (r<0) return -1;
                   Cnt++;
                   Offset += r;
                   if (mx>0)
                       if (Cnt>=mx) break;
              }
              if (Cnt<mn) return -1;
              return Offset;
          }
          if (Cnt<=mn) return -1;
          Cnt--;
          for (int i=0;i<Cnt;i++) {
              Elem.Clear();
              int r = Elem.Match(m,s+Offset,size-Offset,res);
              if (r<0) return -1;
              Offset += r;
          }
          return Offset;
        }        
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res, Int2Type<qtLazy>)
        { unsigned Offset = 0;
          if (mx>0)
              if (Cnt>=mx) return -1;
          for (int i=0;i<Cnt;i++) {
              Elem.Clear();
              int r = Elem.Match(m,s+Offset,size-Offset,res);
              if (r<0) return -1;
              Offset += r;
          }
          for (;;) {
              int r = StopCond(m,s+Offset,size-Offset,Type2Type<U>());
              if (r>=0) break;
              Elem.Clear();
              r = Elem.Match(m,s+Offset,size-Offset,res);
              if (r<=0) break;
              if (Cnt<0) Cnt = 0;
              Cnt++;
              Offset += r;
              if (mx>0)
                  if (Cnt>=mx) break;
          }
          if (Cnt==0) return -1;
          if (Cnt<0)  Cnt = 0;
          if (Cnt<mn) return -1;
          return Offset;
        }        
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res, Int2Type<qtFast>)
        { unsigned Offset = 0;
          if (Cnt>=0) return -1;
          for (Cnt=0;;) {
              int r = StopCond(m,s+Offset,size-Offset,Type2Type<U>());
              if (r>=0) break;
              Elem.Clear();
              r = Elem.Match(m,s+Offset,size-Offset,res);
              if (r<=0) break;
              Cnt++;
              Offset += r;
              if (mx>0)
                  if (Cnt>=mx) break;
          }
          if (Cnt<mn) return -1;
          return Offset;
        }        
        T      Elem;
        U      Stop;
        int    Cnt;
    };

    // 
    //////////////////////////////////////////////////////////////////////////////////////////////
    //    Forget   
    template <class T> class B
	{ public:
        B(): Elem() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { int r = Elem.Match(m,s,size,res);
          if (r>=0) {
              m->Save(s,r);
              if (res!=NULL)
                  res->AddValue(s,r);
          }
          return r;
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) 
        { return Elem.Replace(m,dst,maxsize);
        }
        void Clear(void)
        { Elem.Clear();
        }
      private:
        T Elem;
    };

    // 
    //////////////////////////////////////////////////////////////////////////////////////////////
    // f   
    template <PREDICATE_TYPE p, bool f = false> class P;
    template <bool f> class P<ptSpace,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { int r = -1;
          if (size==0) return -1;
          if (O::Match(m,s,1,res)<0) return -1;
          if (s[0]==' ')  r = 1;
          if (s[0]==0x09) r = 1;
          if (s[0]==0x0A) r = 1;
          if (s[0]==0x0D) r = 1;
          if (f) r = -r;
          return r;
        }
    };
    template <bool f> class P<ptBegin,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { unsigned sz;
          if (O::Match(m,s,0,res)<0) return -1;
          const char* str = m->Catch(0,sz);
          if (str==NULL) return -1;
          int r = -1;
          if (str==s) r = 1;
          if (f) r = -r;
          return r;
        }
    };
    template <bool f> 
	class P<ptEnd,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { if (O::Match(m,s,0,res)<0) return -1;
          if (size==0) return f?-1:0;
          return f?0:-1;
        }
    };
    template <bool f> class P<ptSomeChar,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { if (size==0) return -1;
          if (O::Match(m,s,1,res)<0) return -1;
          if (f) return -1;
          return 1;
        }
    };
    template <bool f> class P<ptDigit,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { int r = -1;
          if (size==0) return -1;
          if (O::Match(m,s,1,res)<0) return -1;
          if ((s[0]>='0')&&(s[0]<='9')) r = 1;
          if (f) r = -r;
          return r;
        }
    };
    template <bool f> class P<ptWordChar,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { int r = -1;
          if (size==0) return -1;
          if (O::Match(m,s,1,res)<0) return -1;
          if (s[0]=='_') r = 1;
          if ((s[0]>='a')&&(s[0]<='z'))         r = 1;
          if ((s[0]>='A')&&(s[0]<='Z'))         r = 1;
          if ((s[0]>='')&&(s[0]<=''))         r = 1;
          if ((s[0]>='')&&(s[0]<=''))         r = 1;
          if ((s[0]=='')||(s[0]==''))         r = 1;
          if (P<ptDigit>().Match(m,s,1,res)==1) r = 1;
          if (f) r = -r;
          return r;
        }
    };
    template <bool f> class P<ptWordBorder,f>: public O
    { public:
        P(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { unsigned sz;
          if (O::Match(m,s,0,res)<0) return -1;
          const char* str = m->Catch(0,sz);
          if (str==NULL) return -1;
          const char* p = s-1;
          if (str==s) {
              if ((size>0)&&(P<ptWordChar>().Match(m,s,1,res)==1)) return f?-1:0;
              return f?0:-1;
          }
          if (size==0) {
              if ((str!=s)&&(P<ptWordChar>().Match(m,p,1,res)==1)) return f?-1:0;
              return f?0:-1;
          }
          if ((P<ptWordChar>().Match(m,p,1,res)==1)&&
              (P<ptWordChar>().Match(m,s,1,res)!=1)) return f?-1:0;
          if ((P<ptWordChar>().Match(m,p,1,res)!=1)&&
              (P<ptWordChar>().Match(m,s,1,res)==1)) return f?-1:0;
          return f?0:-1;
        }
    };

    //  
    //////////////////////////////////////////////////////////////////////////////////////////////
    // f   
    template <int start,int end = 0, bool f = false> 
    class C: public O
    { public:
        C(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { int r = 1;
          if (size==0) return -1;
          if (O::Match(m,s,1,res)<0) return -1;
          if (s[0]<start) r = -1;
          if (end>start) {
              if (s[0]>end) r = -1;
          } else {
              if (s[0]>start) r = -1;
          }
          if (f) r = -r;
          return r;
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) 
        { if (maxsize<1) return esLimitError;
          if ((end!=0)&&(start!=end)) return esOtherError;
          dst[0] = start;
          dst++;
          maxsize--;
          return esNoError;
        }
    };

    //  
    //////////////////////////////////////////////////////////////////////////////////////////////
    // f   
    template <class T, bool f = false> class S;
    template <bool f> class S<NullType,f> 
    { public:
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { return true; 
        } 
    };
    template <class T, class U, bool f>
    class S< TypeList<T,U>, f>: public O
    { public:
        S< TypeList<T,U>, f>(): O(), Head(), Tail() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { int r = -1;
          if (size==0) return -1;
          if (O::Match(m,s,1,res)<0) return -1;
          if ((Head.Match(m,s,1,res)==1)||
              (Tail.Match(m,s,1,res)==1)) r = 1;
          if (f) r = -r;
          return r;
        }
      private:
        T       Head;
        S< U >  Tail;
    };
    template <class T, bool f>
    class S
    { public:
        S(): Elem() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res) 
        { return Elem.Match(m,s,size,res);
        }
      private:
        S<TL_1(T),f> Elem;
    };

    // 
    //////////////////////////////////////////////////////////////////////////////////////////////
    template <int level> 
    class V: public O
    { public:
        V(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { unsigned sz;
          const char* v = m->Catch(level,sz);
          if (v==NULL) return -1;
          if (sz>size) return -1;
          if (memcmp(s,v,sz)==0) return sz;
          return -1;
        }
        ERROR_STATE Replace(CRegExpIf* m, char*& dst, unsigned& maxsize) 
        { unsigned sz;
          const char* v = m->Catch(level);
          if (v==NULL) return esNoMatch;
          if (sz>maxsize) return esLimitError;
          memcpy(dst,v,sz);
          dst += sz;
          maxsize -= sz;
          return esNoError; 
        }
    };

    //   
    //////////////////////////////////////////////////////////////////////////////////////////////
    template <class T = NullType, class U = NullType> class R;
    template <>
    class R<NullType,NullType>: public O
    { public:
        R(): O() {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { if (size==0) return 0;
          std::auto_ptr<CRegExpIf> p(m->Clone());
          return p->Match(res,s,size);
        }
    };
    template <class T, class U>
    class R: public CRegExpIf,
             public O
    { public: 
        R(): CRegExpIf(), Elem(), Found(-1) {}
        int Match(CRegExpIf* m, const char* s, unsigned size, CResIf*& res)
        { Match(res,s,size);
          return Found;
        }
        virtual bool Match(CResIf*& r, const char* src = NULL, unsigned size = 0)
        { CreateNode(r,Type2Type<U>());
          Found = Elem.Match(this,src,size,r);
          return (Found>=0);
        }
        virtual const char* Catch(unsigned level, unsigned& size) {return NULL;}
        virtual ERROR_STATE Replace(char* dst, unsigned size) {return esOtherError;}
        virtual void Reset(void) {}
        virtual CRegExpIf* Clone(void) const { return new R<T,U>; }
      private:
        template <class S>
        void CreateNode(CResIf*& r, Type2Type<S>)
        { CResIf* p = new S;
          if (r!=NULL) r->AddNode(p);
          r = p;
        }
        template <>
        void CreateNode(CResIf*& r, Type2Type<NullType>) {}
        E< T > Elem;
        int    Found;
    };

    //   
    //////////////////////////////////////////////////////////////////////////////////////////////
    template < class T, class U = NullType>
    class CRegExp: public CRegExpIf
    { public:
        CRegExp(): Expr(), Repl(), Start(NULL), Str(NULL), Size(0), Found(-1) {}

        virtual bool           Match(CResIf*& r, const char* src = NULL, unsigned size = 0);
        virtual const char*    Catch(unsigned level, unsigned& size);
        virtual ERROR_STATE  Replace(char* dst, unsigned size);
        virtual void           Reset(void) { CRegExpIf::Reset(); Found = -1; }
        virtual CRegExpIf*     Clone(void) const { return new CRegExp<T,U>; }

      private:
        E< T >       Expr;
        E< U >       Repl;
        const char*  Start;
        const char*  Str;
        unsigned     Size;

	  public:
        int          Found;
    };

    template <class T, class U>
    bool CRegExp<T,U>::Match(CResIf*& r, const char* src, unsigned size) 
    { 
      if (src!=NULL) {
          Str   = src;
          Start = src;    
          Size  = size;
          if (Size==0) 
              Size = (unsigned)strlen(src);
      } else {
          if (Str==NULL) return false;
          int step = (Found<=0)?1:Found;
          if (Size<=(unsigned)step) return false;
          Str  += step;
          Size -= step;
      }
      Reset();
      while ((Found=Expr.Match(this,Str,Size,r))==-1) {
          if (Size<=1) return false;
          Str++;
          Size--;
      }
      return true;
    }

    template <class T, class U>
    ERROR_STATE CRegExp<T,U>::Replace(char* dst, unsigned size)
    { return Repl.Replace(this,dst,size);
    }

    template <class T, class U>
    const char* CRegExp<T,U>::Catch(unsigned level, unsigned& size)
    {
      if (level==0) {
          size = (Found<0)?Size:Found;
          return (Found<0)?Start:Str;
      }
      return CRegExpIf::Catch(level,size);
    }

}